----------------------------------------------------------------------------
-------------------------      M a R T E   O S      ------------------------
----------------------------------------------------------------------------
--                                                           {MARTE_VERSION}
--
--                       'r t _ e p _ p r o t o c o l . a d s'
--
--                                     Ada
--
--
--  File 'rt_ep_protocol.ads'                                        By Chema.
--                                                        Jose Maria Martinez
--                                                       <martinjm@unican.es>
--  Import declarations of rt_comm.h
--
--
--  {MARTE_COPYRIGHT}
--
-----------------------------------------------------------------------------
--  {<MAR}
--  Chema:
--  Date format DD-MM-YY:
--
--
--  15-09-03: First FINAL draft (First attempt was in 2002).
--
--
--  {MAR>} -------------------------------------------------------------------

--  with Ada.Streams;
with Interfaces;

package Rt_Ep_Protocol  is

   Max_Info_Size : constant := 1492;
   type Ethernet_Address is array (1 .. 6) of Interfaces.Unsigned_8;
   for Ethernet_Address'Size use 48;

   subtype RT_EP_Priority is Interfaces.Unsigned_8
     range 1 .. Interfaces.Unsigned_8'Last;

   subtype RT_EP_Channel_Number is Interfaces.Unsigned_16
     range 1 .. Interfaces.Unsigned_16'Last;

   subtype RT_EP_Packet_Size is Interfaces.Unsigned_16
     range 1 .. Max_Info_Size;

   --  Exceptions :
   --  If the station is no longer on the ring.
   Station_Not_Valid : exception;
   --  If the station isn't on the logical ring
   Station_Not_Found : exception;
   --  If the channel is not available
   Invalid_Channel : exception;
   --  If a synchronization queue error
   Synch_Error : exception;
   --  If not being able creating the queues
   Creation_Error : exception;
   --  If an unknown error has occurred
   Unexpected_Error : exception;

   ---------------
   --  Send_Info --
   ---------------

   --  Send_Info is used to send Data over the network.
   --  You have to provide the destination MAC addres (Destination_Address)
   --  the channel identifier of the reception task (Channel_ID)
   --  the priority of the packet.
   --  and the length of the data in bytes.

   --  Raise on Error:
   --  Station_Not_Valid
   --  Station_Not_Found
   --  Invalid_Channel
   --  Synch_Error
   --  Unexpected_Error



   -----------------------
   --  Generic Send Info --
   -----------------------
   generic
      type Data_Type is private;
   procedure Generic_Send_Info
     (Destination_Address : in out Ethernet_Address;
      Channel_ID : in RT_EP_Channel_Number;
      Data : in Data_Type;
      Size : in RT_EP_Packet_Size;
      Data_Priority : in RT_EP_Priority);

   ---------------
   --  Recv_Info --
   ---------------
   --  Recv_Info reads from the Chanel_ID the highest information Data
   --  and store it in Data. The length is stored in Size in bytes and
   --  the priority of the message is stored in Data_Priority.

   --  Raise on Error:
   --  Invalid_Channel
   --  Synch_Error
   --  Unexpected_Error

   --  If error, the value stored in Length_Data will be zero or less.
   --  The size is in bytes.
   -----------------------
   --  Generic Recv_Info --
   -----------------------
   generic
      type Data_Type is private;
   procedure Generic_Recv_Info
     (Source_Address : in out Ethernet_Address;
      Channel_ID : in RT_EP_Channel_Number;
      Data : out Data_Type;
      Size : out RT_EP_Packet_Size;
      Data_Priority : out RT_EP_Priority);

   --------------
   --  Any_Info --
   --------------

   --  Any_Info checks if there is any data to be received in the specified
   --  Channel_ID. The caller process MUST assign its ID to
   --  the channel_id variable. Returns: True if threre is Data to be received.

   --  Raise on Error:
   --  Invalid_Channel
   --  Synch_Error
   --  Unexpected_Error
   function Any_Info (Channel_ID : in RT_EP_Channel_Number) return Boolean;

   ---------------
   --  Init_Comm --
   ---------------
   --  init_comm must be used to start the communications
   --  thread and the tx and rx queues.
   --  Returns 0 if OK.

   --  Raise on Error:
   --  Creation_Error
   --  Unexpected_Error

   function Init_Comm (Prio_Ceiling : Integer) return Integer;








--     -------------------------------------------------------------------
--     --- RT-EP with Streams  -------------------------------------------
--     -------------------------------------------------------------------

--     type Function_Stub_type is abstract
--       new Ada.Streams.Root_Stream_Type with private;
--  typeFunction_Stub_Type_Ac is access all Ada.Streams.Root_Stream_Type'Class;

--     type Function_ID_Type is new Interfaces.Unsigned_8;

--     --  This function will create an access to a Stream type.
--     procedure Stream_Stub (Function_Stub : in out  Function_Stub_Type;
--                            Function_Ptr : out Function_Stub_Type_Ac);

--     procedure Put_Function_ID (Stream : in out Function_Stub_Type;
--                               Function_ID : in Function_ID_Type);

--     function Get_Function_ID (Stream : in Function_Stub_Type)
--                             return Function_ID_Type;









--     ---------------
--     --  Recv_Info --
--     ---------------

--     --  Raise on Error:
--     --  Invalid_Channel
--     --  Synch_Error
--     --  Unexpected_Error

--     procedure Recv_Info (Source_Address : in out Ethernet_Address;
--                          Channel_ID : in RT_EP_Channel_Number;
--                          Data : in out Function_Stub_Type;
--                          Size : out RT_EP_Packet_Size;
--                          Data_Priority : out RT_EP_Priority);
--     ---------------
--     --  Send_Info --
--     ---------------

--     --  Raise on Error:
--     --  Station_Not_Valid
--     --  Station_Not_Found
--     --  Invalid_Channel
--     --  Synch_Error
--     --  Unexpected_Error

--     procedure Send_Info (Destination_Address : in out Ethernet_Address;
--                          Channel_ID : in RT_EP_Channel_Number;
--                          Data : in out Function_Stub_Type;
--                          Size : in RT_EP_Packet_Size;
--                          Data_Priority : in RT_EP_Priority);

--  private


--     type Function_Stub_type is abstract
--       new Ada.Streams.Root_Stream_Type with record
--          Function_ID : Function_ID_Type;
--       end record;

end Rt_Ep_Protocol;

