/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                  'p r i o _ q u e u e _ m o n i t o r . c'
 *
 *                                     C
 *
 *                                               
 * File 'prio_queue_monitor.c'                                        By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Queue access monitor implementation (Body functions of prio_queue_monitor.h)
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

/* prio_queue_monitor.c -> Object Implementation */

#include <sched.h>
#include <pthread.h>
#include "prio_queue.h"
#include "prio_queue_monitor.h"


static pthread_mutex_t the_mutex[MAX_CHANNELS+1];
static pthread_cond_t the_condition[MAX_CHANNELS+1];

int create_queue(int prio_ceiling, queue_t *Q, const int queue_num){ 

  pthread_mutexattr_t attr;

  /*Inicialize the mutex*/

  if(pthread_mutexattr_init(&attr)!=0)
    return CREATION_ERROR;
  if(pthread_mutexattr_setprotocol(&attr,PTHREAD_PRIO_PROTECT)!=0)
    return CREATION_ERROR;
  if(pthread_mutexattr_setprioceiling(&attr,prio_ceiling)!=0)
    return CREATION_ERROR;
  if(pthread_mutex_init(&the_mutex[queue_num],&attr)!=0)
    return CREATION_ERROR;
  if(pthread_mutexattr_destroy(&attr)!=0)
    return CREATION_ERROR;

   /*Inicialize the cond vatiable with default attributes*/
  if(pthread_cond_init(&the_condition[queue_num],NULL)!=0)
    return CREATION_ERROR;
  /* Inicialize de queue*/
  init_queue(Q);
  return 0;
} //end function create_queue 


int queue_insert(element_t *E, priority_t P,queue_t *Q, const int queue_num){ 

  if(pthread_mutex_lock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;
  enqueue(E,P,Q);
  if(pthread_cond_signal(&the_condition[queue_num])!=0)
    return SYNCH_ERROR;
#ifdef MONITOR_DEBUG_MODE
  printf("prio_queue_monitor:queue_insert>Signal to queue %d sended\n",
	 queue_num);
#endif
  if(pthread_mutex_unlock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;
  return 0;

}//end function queue_insert

int queue_extract(element_t **E,priority_t *P,queue_t *Q, 
		  const int queue_num){ 

  if(pthread_mutex_lock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

  while(empty(Q)){
#ifdef MONITOR_DEBUG_MODE
    printf("prio_queue_monitor:queue_extract>The queue %d is empty."
	   "We begin to wait.\n", queue_num);
#endif
    if(pthread_cond_wait(&the_condition[queue_num],&the_mutex[queue_num])!=0)
      return SYNCH_ERROR;
#ifdef MONITOR_DEBUG_MODE
    printf("prio_queue_monitor:queue_extract>"
	   "The queue %d is with one element.\n",
	   queue_num);
#endif

  } //end while.

  dequeue(E,P,Q);

  if(pthread_mutex_unlock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

  return 0;

}//end function queue_extract


int read_first_in_queue(element_t **E, priority_t *P, queue_t *Q, 
			const int queue_num){

  if(pthread_mutex_lock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

#ifdef MONITOR_DEBUG_MODE
    printf("prio_queue_monitor:read_first_in_queue>Reading the queue %d.\n",
	   queue_num);
#endif

  if(empty(Q)){
     if(pthread_mutex_unlock(&the_mutex[queue_num])!=0)
       return SYNCH_ERROR;

    return QUEUE_EMPTY;
  }
  
  read_first(E,P,Q);

  if(pthread_mutex_unlock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

  return 0;


}//end function read_first_in_queue



int is_queue_empty(element_t *E,queue_t *Q,const int queue_num){

  if(pthread_mutex_lock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

  if(empty(Q)){
     if(pthread_mutex_unlock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

    return 1;
  }
  
  if(pthread_mutex_unlock(&the_mutex[queue_num])!=0)
    return SYNCH_ERROR;

  return 0;

}//end function is_queue_empty
